#include <ruby.h>
#include <math.h>
#include <narray.h>


#define cNArray rb_const_get(ruby_class, rb_intern("NArray"))

extern VALUE ruby_class;

static int ssl2_arysize(x)
     VALUE x;
{
  int n;
  int na;
  int i;
  VALUE *ptr;

  if ( TYPE(x)!=T_ARRAY )
    rb_raise(rb_eTypeError, "x is not array");

  n = RARRAY(x)->len;
  na = n;
  ptr = RARRAY(x)->ptr;
  for ( i=0; i<n; i++ ) {
    if ( TYPE(ptr[i])==T_ARRAY )
      na += ssl2_arysize(ptr[i])-1;
  }
  return na;
}

static int ssl2_aryrank(x)
     VALUE x;
{
  int rank;
  VALUE *ptr;
  int size;
  int n;

  if (TYPE(x)!=T_ARRAY)
    rb_raise(rb_eTypeError, "x in not array");

  n = RARRAY(x)->len;
  ptr = RARRAY(x)->ptr;
  if (TYPE(ptr[0])==T_ARRAY) {
    rank = ssl2_aryrank(ptr[0])+1;
  } else
    rank = 1;

  return rank;
}

static void ssl2_aryshape(x,shape)
     VALUE x; int* shape;
{
  int n;
  VALUE *ptr;

  n = RARRAY(x)->len;
  ptr = RARRAY(x)->ptr;
  shape[0] = n;
  if (TYPE(ptr[0])==T_ARRAY)
      ssl2_aryshape(ptr[0],shape[1]);
}  

static void ssl2_ary_check_form(x)
     VALUE x;
{
  int n;
  VALUE *ptr;
  int size, rank;

  int i;

  n = RARRAY(x)->len;
  ptr = RARRAY(x)->ptr;
  if (TYPE(ptr[0])==T_ARRAY) {
    size = ssl2_arysize(ptr[0]);
    rank = ssl2_aryrank(ptr[0]);
    for (i=1;i<n;i++){
      if (TYPE(ptr[i])!=T_ARRAY)
	rb_raise(rb_eTypeError, "x is not well formed array (type)");
      if (ssl2_arysize(ptr[i])!=size)
	rb_raise(rb_eRuntimeError, "x is not well formed array (size)");
      if (ssl2_aryrank(ptr[i])!=rank)
	rb_raise(rb_eRuntimeError, "x is not well formed array (rank)");
    }
    for (i=0;i<n;i++)
      ssl2_ary_check_form(ptr[i]);
  }
}

static float *ssl2_arysetvalue(x,cx)
     VALUE x; float *cx;
{
  VALUE *ptr;
  int n;
  int i;

  if ( TYPE(x)!=T_ARRAY )
    rb_raise(rb_eTypeError, "x is not array");

  n = RARRAY(x)->len;
  ptr = RARRAY(x)->ptr;
  for ( i=0; i<n; i++ ) {
    if ( TYPE(ptr[i])==T_ARRAY )
      cx = ssl2_arysetvalue(ptr[i],cx);
    else if ( TYPE(rb_Float(ptr[i]))==T_FLOAT ) {
      *cx = (float) RFLOAT(rb_Float(ptr[i]))->value;
      cx++;
    }
    else
      rb_raise(rb_eTypeError, "value is invalid type");
  }
  return cx;
}


static int *ssl2_arysetvaluei(x,cx)
     VALUE x; int *cx;
{
  VALUE *ptr;
  int n;
  int i;

  if ( TYPE(x)!=T_ARRAY )
    rb_raise(rb_eTypeError, "x is not array");

  n = RARRAY(x)->len;
  ptr = RARRAY(x)->ptr;
  for ( i=0; i<n; i++ ) {
    if ( TYPE(ptr[i])==T_ARRAY )
      cx = ssl2_arysetvaluei(ptr[i],cx);
    else if ( TYPE(rb_Integer(ptr[i]))==T_FIXNUM || TYPE(rb_Integer(ptr[i]))==T_BIGNUM ) {
      *cx = (int)NUM2INT(rb_Integer(ptr[i]));
      cx++;
    }
    else
      rb_raise(rb_eTypeError, "value is invalid type");
  }
  return cx;
}


static float *ssl2_ary2cary(x,pn)
     VALUE x; int *pn;
{
  float *cx;
  int n;
  int i,j;
  VALUE *ptr;

  n = ssl2_arysize(x);
  *pn = n;
  cx = (float *) ALLOC_N(float,n);
  ssl2_arysetvalue(x,cx);
  return cx;
}

static int *ssl2_ary2caryi(x,pn)
     VALUE x; int *pn;
{
  int *cx;
  int n;
  int i,j;
  VALUE *ptr;

  n = ssl2_arysize(x);
  *pn = n;
  cx = (int *) ALLOC_N(int,n);
  ssl2_arysetvaluei(x,cx);
  return cx;
}


static float *ssl2_ary2cary_withshape(x,shape,rank)
     VALUE x; int *shape, *rank;
{
  float *cx;
  int *pn;

  ssl2_ary_check_form(x);
  *rank = ssl2_aryrank(x);
  shape = ALLOC_N(int,*rank);
  ssl2_aryshape(x,shape);
  cx = ssl2_ary2cary(x,pn);
  return cx;
}


#ifdef HAVE_NARRAY_H
static float *ssl2_nary2cary(nx,pn)
     VALUE nx; int *pn;
{
  float *cx;
  struct NARRAY *nax;
  float *ptr;
  int n;
  int i;

  nx = na_cast_object(nx, NA_SFLOAT);
  GetNArray(nx,nax);
  n = nax->total; 
  *pn = n;
  ptr = (float *)NA_PTR(nax,0);
  cx = ALLOC_N(float,n);
  for ( i=0; i<n; i++ ) {
    cx[i] = ptr[i];
  }
  return cx;
}


static int *ssl2_nary2caryi(nx,pn)
     VALUE nx; int *pn;
{
  int *cx;
  struct NARRAY *nax;
  int *ptr;
  int n;
  int i;

  nx = na_cast_object(nx, NA_SINT);
  GetNArray(nx,nax);
  n = nax->total; 
  *pn = n;
  ptr = (int *)NA_PTR(nax,0);
  cx = ALLOC_N(int,n);
  for ( i=0; i<n; i++ ) {
    cx[i] = ptr[i];
  }
  return cx;
}


static float *ssl2_nary2cary_withshape(nx,shape,rank)
     VALUE nx; int *shape, *rank;
{
  float *cx;
  int *pn;
  struct NARRAY *nax;
  float *ptr;

  nx = na_cast_object(nx, NA_SFLOAT);
  GetNArray(nx,nax);
  *rank = nax->rank;
  shape = nax->shape;
  cx = ssl2_nary2cary(nx,pn);
  return cx;
}
#endif


static VALUE ssl2_cary2ary(cx,dim,shape)
     float **cx; int dim; int *shape;
{
  VALUE ary;
  int i;
  if ( dim==0 ){
    ary = rb_float_new((double)*(*cx)++);
  }
  else {
    ary = rb_ary_new2(shape[dim-1]);
    for ( i=0;i<shape[dim-1];i++ )
      RARRAY(ary)->ptr[i] = ssl2_cary2ary(cx,dim-1,shape);
    RARRAY(ary)->len = shape[dim-1];
  }
  return ary;
}


static VALUE ssl2_cary2aryi(cx,dim,shape)
     int **cx; int dim; int *shape;
{
  VALUE ary;
  int i;
  if ( dim==0 ){
    ary = NUM2INT(*(*cx)++);
  }
  else {
    ary = rb_ary_new2(shape[dim-1]);
    for ( i=0;i<shape[dim-1];i++ )
      RARRAY(ary)->ptr[i] = ssl2_cary2aryi(cx,dim-1,shape);
    RARRAY(ary)->len = shape[dim-1];
  }
  return ary;
}


static VALUE ssl2_cary2aryl(cx,dim,shape)
     long **cx; int dim; int *shape;
{
  VALUE ary;
  int i;
  if ( dim==0 ){
    ary = NUM2LONG(*(*cx)++);
  }
  else {
    ary = rb_ary_new2(shape[dim-1]);
    for ( i=0;i<shape[dim-1];i++ )
      RARRAY(ary)->ptr[i] = ssl2_cary2aryl(cx,dim-1,shape);
    RARRAY(ary)->len = shape[dim-1];
  }
  return ary;
}


#ifdef HAVE_NARRAY_H
static VALUE ssl2_cary2nary(cx,dim,shape)
     float *cx; int dim; int *shape;
{
  float *ptr;
  int total;
  struct NARRAY *nax;
  VALUE nx;
  int i;

  nx = na_make_object(NA_SFLOAT,dim,shape,cNArray);

  total = 1;
  for ( i=0; i<dim; i++ )
    total *= shape[i];

  GetNArray(nx,nax);
  ptr = (float *) NA_PTR(nax,0);
  for ( i=0; i<total; i++ ) {
    ptr[i] = cx[i];
  }
  return nx;
}


static VALUE ssl2_cary2naryi(cx,dim,shape)
     int *cx; int dim; int *shape;
{
  int *ptr;
  int total;
  struct NARRAY *nax;
  VALUE nx;
  int i;

  nx = na_make_object(NA_LINT,dim,shape,cNArray);

  total = 1;
  for ( i=0; i<dim; i++ )
    total *= shape[i];

  GetNArray(nx,nax);
  ptr = (int *) NA_PTR(nax,0);
  for ( i=0; i<total; i++ ) {
    ptr[i] = cx[i];
  }
  return nx;
}


static VALUE ssl2_cary2naryl(cx,dim,shape)
     long *cx; int dim; int *shape;
{
  long *ptr;
  int total;
  struct NARRAY *nax;
  VALUE nx;
  int i;

  nx = na_make_object(NA_SINT,dim,shape,cNArray);

  total = 1;
  for ( i=0; i<dim; i++ )
    total *= shape[i];

  GetNArray(nx,nax);
  ptr = (long *) NA_PTR(nax,0);
  for ( i=0; i<total; i++ ) {
    ptr[i] = cx[i];
  }
  return nx;
}
#endif


float *ssl2_getcary(x,pn)
     VALUE x; int *pn;
{
  float *cx;

  if (TYPE(x) == T_FLOAT)
    x = rb_Array(x);

  if ( TYPE(x) == T_ARRAY )
    cx = ssl2_ary2cary(x,pn);
#ifdef HAVE_NARRAY_H
  else if ( rb_obj_is_kind_of(x,cNArray)== Qtrue)
    cx = ssl2_nary2cary(x,pn);
#endif
  else
    rb_raise(rb_eTypeError, "x is invalid type");
  return cx;
}


int *ssl2_getcaryi(x,pn)
     VALUE x; int *pn;
{
  int *cx;

  if ( TYPE(x)==T_FIXNUM || TYPE(x)==T_BIGNUM )
    x = rb_Array(x);

  if ( TYPE(x)==T_ARRAY )
    cx = ssl2_ary2caryi(x,pn);
#ifdef HAVE_NARRAY_H
  else if ( rb_obj_is_kind_of(x,cNArray)== Qtrue)
    cx = ssl2_nary2caryi(x,pn);
#endif
  else
    rb_raise(rb_eTypeError, "x is invalid type");
  return cx;
}


float *ssl2_getcary_withshape(x,shape,rank)
     VALUE x; int *shape, *rank;
{
  float *cx;
  int *pn;

  if (TYPE(x) == T_FLOAT)
    x = rb_Array(x);

  if ( TYPE(x) == T_ARRAY )
    cx = ssl2_ary2cary_withshape(x,shape,rank);
#ifdef HAVE_NARRAY_H
  else if ( rb_obj_is_kind_of(x,cNArray)== Qtrue)
    cx = ssl2_nary2cary_withshape(x,shape,rank);
#endif
  else
    rb_raise(rb_eTypeError, "x is invalid type");
  return cx;
}


VALUE ssl2_getary(cx,dim,shape)
     float *cx; int dim; int *shape;
{
#ifdef HAVE_NARRAY_H
  return ssl2_cary2nary(cx,dim,shape);
#else
  return ssl2_cary2ary(&cx,dim,shape);
#endif
}


VALUE ssl2_getaryi(cx,dim,shape)
     int *cx; int dim; int *shape;
{
#ifdef HAVE_NARRAY_H
  return ssl2_cary2naryi(cx,dim,shape);
#else
  return ssl2_cary2aryi(&cx,dim,shape);
#endif
}


float func(cx)
     float *cx;
{
  return (float)NUM2DBL(rb_yield(rb_float_new((double)*cx)));
}

void sub(cx,cy)
     float *cx, *cy;
{
  VALUE y;
  int m;
  y = rb_yield(rb_float_new((double)*cx));
  cy = ssl2_getcary(y,&m);
}
