#include <ruby.h>
#include <math.h>
#include <stdio.h>
#include <narray.h>
#include "ssl2.h"

extern void eig1_(float*,int*,int*,int*,float*,float*,float*,float*,int*);
extern void ceig2_(float*,int*,int*,int*,float*,float*,float*,int*,int*);
extern void seig1_(float*,int*,float*,float*,int*,int*,float*,int*);
extern void seig2_(float*,int*,int*,float*,float*,int*,float*,int*);
extern void heig2_(float*,int*,int*,int*,float*,float*,float*,float*,int*);
extern void bseg_(float*,int*,int*,int*,int*,float*,float*,float*,int*,float*,int*);
extern void bsegj_(float*,int*,int*,int*,float*,int*,float*,float*,int*,int*,float*,int*);
extern void teig1_(float*,float*,int*,float*,float*,int*,int*,int*);
extern void teig2_(float*,float*,int*,int*,float*,float*,int*,float*,int*);
extern void gseg2_(float*,float*,int*,int*,float*,float*,float*,float*,int*,float*,int*);
extern void gbseg_(float*,float*,int*,int*,int*,float*,float*,int*,float*,float*,int*,int*,float*,int*);

extern void hsqr_();
extern void chsqr_();
extern void trql_();
extern void bsct1_();

extern void hvec_();
extern void chvec_();
extern void bsvec_();

extern void blnc_();
extern void cblnc_();
extern void hes1_();
extern void ches2_();
extern void trid1_();
extern void tridh_();
extern void btrid_();
extern void hbk1_();
extern void chbk2_();
extern void trbk_();
extern void trbkh_();
extern void nrml_();
extern void cnrml_();
extern void gschl_();
extern void gsbk_();


static VALUE
rb_eig1(argc, argv, self)
     int argc; VALUE *argv, self;
{
  float *ca;
  int n;
  int cmode;
  float *er, *ei;
  float *ev;
  float *vw;
  int icon;

  VALUE a, mode;
  VALUE ner,nei,nev;
  int *shape;

  int nn;

  rb_scan_args(argc, argv, "11", &a, &mode);
  if ( mode==Qnil )
    cmode = 0;
  else {
    switch ( TYPE(mode) ) {
    case T_FIXNUM:
      cmode = NUM2INT(mode);
      break;
    case T_TRUE:
      cmode = 1;
      break;
    case T_FALSE:
      cmode = 0;
      break;
    default:
      rb_raise(rb_eTypeError, "mode is invalid type");
      break;
    }
  }

  ca = ssl2_getcary(a,&nn);

  n = sqrt(nn);
  if ( nn != n*n )
    rb_raise(rb_eRuntimeError, "a.length != n*n");

  er = ALLOC_N(float,n);
  ei = ALLOC_N(float,n);
  ev = ALLOC_N(float,n*n);
  vw = ALLOC_N(float,n);


  eig1_(ca,&n,&n,&cmode,er,ei,ev,vw,&icon);
  free(vw);
  free(ca);

  shape = ALLOC_N(int,1);
  shape[0] = n;
  ner = ssl2_getary(er,1,shape);
  nei = ssl2_getary(ei,1,shape);
  free(shape);
  shape = ALLOC_N(int,2);
  shape[0] = n;
  shape[1] = n;
  nev = ssl2_getary(ev,2,shape);
  free(shape);

  free(er);
  free(ei);
  free(ev);

  if ( icon!=0 )
    fprintf(stderr,"icon is %d\n",icon);
  return rb_ary_new3(3,ner,nei,nev);
}
 

static VALUE
rb_ceig2(argc, argv, self)
     int argc; VALUE *argv, self;
{
  float *cza;
  int n;
  int cmode;
  float *ze;
  float *zev;
  float *vw;
  int *ivw;
  int icon;

  VALUE za, mode;
  VALUE nze,nzev;
  int *shape;

  int nn;

  rb_scan_args(argc, argv, "11", &za, &mode);
  if ( mode==Qnil )
    cmode = 0;
  else {
    switch ( TYPE(mode) ) {
    case T_FIXNUM:
      cmode = NUM2INT(mode);
      break;
    case T_TRUE:
      cmode = 1;
      break;
    case T_FALSE:
      cmode = 0;
      break;
    default:
      rb_raise(rb_eTypeError, "mode is invalid type");
      break;
    }
  }
  cza = ssl2_getcary(za,&nn);

  n = sqrt(nn/2);
  if ( nn != n*n*2 )
    rb_raise(rb_eRuntimeError, "za.length != 2*n*n");


  ze = ALLOC_N(float,n*2);
  zev = ALLOC_N(float,n*n*2);
  vw = ALLOC_N(float,n);
  ivw = ALLOC_N(int,n);

  ceig2_(cza,&n,&n,&cmode,ze,zev,vw,ivw,&icon);
  free(cza);
  free(vw);
  free(ivw);

  shape = ALLOC_N(int,2);
  shape[0] = 2;
  shape[1] = n;
  nze = ssl2_getary(ze,2,shape);
  free(shape);
  shape = ALLOC_N(int,3);
  shape[0] = 2;
  shape[1] = n;
  shape[2] = n;
  nzev = ssl2_getary(zev,3,shape);
  free(shape);

  free(ze);
  free(zev);

  if ( icon!=0 )
    fprintf(stderr,"icon is %d\n",icon);
  return rb_ary_new3(2,nze,nzev);
}
 

static VALUE
rb_seig1(self, a)
     VALUE self, a;
{
  float *ca;
  int n;
  float *e;
  float *ev;
  int m;
  float *vw;
  int icon;

  VALUE ne,nev;
  int *shape;

  int nn;

  ca = ssl2_getcary(a,&nn);

  n = (sqrt(8*nn+1)-1)/2;
  if ( nn != n*(n+1)/2 )
    rb_raise(rb_eRuntimeError, "a.length != n(n+1)/2");

  e = ALLOC_N(float,n); 
  ev = ALLOC_N(float,n*n);
  vw = ALLOC_N(float,2*n);

  seig1_(ca,&n,e,ev,&n,&m,vw,&icon);
  free(vw);
  free(ca);

  shape = ALLOC_N(int,1);
  shape[0] = m;
  ne = ssl2_getary(e,1,shape);
  free(shape);
  shape = ALLOC_N(int,2);
  shape[0] = n;
  shape[1] = m;
  nev = ssl2_getary(ev,2,shape);
  free(shape);

  free(e);
  free(ev);

  if ( icon!=0 )
    fprintf(stderr,"icon is %d\n",icon);
  return rb_ary_new3(3,ne,nev,INT2NUM(m));
}
 

static VALUE
rb_seig2(self, a, m)
     VALUE self, a, m;
{
  float *ca;
  int n;
  int cm;
  float *e;
  float *ev;
  float *vw;
  int icon;

  VALUE ne,nev;
  int *shape;

  int nn;

  cm = NUM2INT(m);

  ca = ssl2_getcary(a,&nn);

  n = (sqrt(8*nn+1)-1)/2;
  if ( nn != n*(n+1)/2 )
    rb_raise(rb_eRuntimeError, "a.length != n(n+1)/2");

  if ( abs(cm)<1 )
    rb_raise(rb_eRuntimeError, "abs(m) must be >= 1");
  else if ( abs(cm)>n )
    rb_raise(rb_eRuntimeError, "abs(m) must be <= n");

  e = ALLOC_N(float,abs(cm)); 
  ev = ALLOC_N(float,n*abs(cm));
  vw = ALLOC_N(float,7*n);

  seig2_(ca,&n,&cm,e,ev,&n,vw,&icon);
  free(vw);
  free(ca);

  shape = ALLOC_N(int,1);
  shape[0] = abs(cm);
  ne = ssl2_getary(e,1,shape);
  free(shape);
  shape = ALLOC_N(int,2);
  shape[0] = n;
  shape[1] = abs(cm);
  nev = ssl2_getary(ev,2,shape);
  free(shape);

  free(e);
  free(ev);

  if ( icon!=0 )
    fprintf(stderr,"icon is %d\n",icon);
  return rb_ary_new3(2,ne,nev);
}


static VALUE
rb_heig2(self, a, m)
     VALUE self, a, m;
{
  float *ca;
  int n;
  int cm;
  float *e;
  float *evr, *evi;
  float *vw;
  int icon;

  VALUE ne;
  VALUE nevr, nevi;
  int *shape;

  int nn;

  cm = NUM2INT(m);

  ca = ssl2_getcary(a,&nn);

  n = sqrt(nn);
  if ( nn != n*n )
    rb_raise(rb_eRuntimeError, "a.length != n*n");

  if ( abs(cm)<1 )
    rb_raise(rb_eRuntimeError, "abs(m) must be >= 1");
  else if ( abs(cm)>n )
    rb_raise(rb_eRuntimeError, "abs(m) must be <= n");

  e = ALLOC_N(float,abs(cm)); 
  evr = ALLOC_N(float,n*abs(cm));
  evi = ALLOC_N(float,n*abs(cm));
  vw = ALLOC_N(float,9*n);

  heig2_(ca,&n,&n,&cm,e,evr,evi,vw,&icon);
  free(vw);
  free(ca);

  shape = ALLOC_N(int,1);
  shape[0] = abs(cm);
  ne = ssl2_getary(e,1,shape);
  free(shape);
  shape = ALLOC_N(int,2);
  shape[0] = n;
  shape[1] = abs(cm);
  nevr = ssl2_getary(evr,2,shape);
  nevi = ssl2_getary(evi,2,shape);
  free(shape);

  free(e);
  free(evr);
  free(evi);

  if ( icon!=0 )
    fprintf(stderr,"icon is %d\n",icon);
  return rb_ary_new3(3,ne,nevr,nevi);
}


static VALUE
rb_bseg(self, a, nh, m, nv, epst)
     VALUE self, a, nh, m, nv, epst;
{
  float *ca;
  int n;
  int cnh;
  int cm;
  int cnv;
  float cepst;
  float *e;
  float *ev;
  float *vw;
  int icon;

  VALUE ne;
  VALUE nev;
  int *shape;

  int nn;

  cnh = NUM2INT(nh);
  cm = NUM2INT(m);
  cnv = NUM2INT(nv);
  cepst = (float) NUM2DBL(epst);

  ca = ssl2_getcary(a,&nn);

  n = nn/(cnh+1)+cnh/2;
  if ( nn != n*(cnh+1)-cnh*(cnh+1)/2 )
    rb_raise(rb_eRuntimeError, "a.length != n*(nh+1)-nh*(nh+1)/2");

  if ( abs(cm)<1 )
    rb_raise(rb_eRuntimeError, "abs(m) must be >= 1");
  else if ( abs(cm)>n ) 
    rb_raise(rb_eRuntimeError, "abs(m) must be <= n");

  if ( cnv<0 )
    rb_raise(rb_eRuntimeError, "nv must be >= 0");
  else if ( cnv>m )
    rb_raise(rb_eRuntimeError, "nv must be <= m");

  if ( cnh<0 )
    rb_raise(rb_eRuntimeError, "nh must be >= 0");
  else if ( cnh>=n )
    rb_raise(rb_eRuntimeError, "nh must be < n");

  e = ALLOC_N(float,abs(cm)); 
  ev = ALLOC_N(float,n*cnv);
  vw = ALLOC_N(float,MAX(3*n+2*cm,2*n*(cnh+1)));

  bseg_(ca,&n,&cnh,&cm,&cnv,&cepst,e,ev,&n,vw,&icon);
  free(vw);
  free(ca);

  shape = ALLOC_N(int,1);
  shape[0] = abs(cm);
  ne = ssl2_getary(e,1,shape);
  free(shape);
  shape = ALLOC_N(int,2);
  shape[0] = n;
  shape[1] = cnv;
  nev = ssl2_getary(ev,2,shape);
  free(shape);

  free(e);
  free(ev);

  if ( icon!=0 )
    fprintf(stderr,"icon is %d\n",icon);
  return rb_ary_new3(2,ne,nev);
}


static VALUE
rb_bsegj(self, a, nh, m, epst, lm, ev)
     VALUE self, a, nh, m, epst, lm, ev;
{
  float *ca;
  int n;
  int cnh;
  int cm;
  float cepst;
  int clm;
  float *e;
  float *cev;
  float *vw;
  int icon;

  VALUE ne;
  VALUE nev;
  int *shape;
  int it;

  int nn;

  cnh = NUM2INT(nh);
  cm = NUM2INT(m);
  cepst = (float) NUM2DBL(epst);
  clm = NUM2INT(lm);

  ca = ssl2_getcary(a,&nn);
  n = nn/(cnh+1)+cnh/2;
  if ( nn != n*(cnh+1)-cnh*(cnh+1)/2 )
    rb_raise(rb_eRuntimeError, "a.length != n*(nh+1)-nh*(nh+1)/2");

  cev = ssl2_getcary(ev,&nn);
  if ( nn!=n*cm )
    rb_raise(rb_eRuntimeError, "ve.length != n*cm");

  if ( abs(cm)<1 )
    rb_raise(rb_eRuntimeError, "abs(m) must be >= 1");
  else if ( abs(cm)>n )
    rb_raise(rb_eRuntimeError, "abs(m) must be <= n");

  if ( cnh<0 )
    rb_raise(rb_eRuntimeError, "nh must be >= 0");
  else if ( cnh>=n )
    rb_raise(rb_eRuntimeError, "nh must be < n");

  if ( clm<1 )
    rb_raise(rb_eRuntimeError, "lm must be > 0");

  e = ALLOC_N(float,abs(cm));
  REALLOC_N(cev,float,n*(abs(cm)+2));
  vw = ALLOC_N(float,MAX(n,2*cm)+cm*(3*cm+1)/2);

  bsegj_(ca,&n,&cnh,&cm,&cepst,&clm,e,cev,&n,&it,vw,&icon);
  free(vw);
  free(ca);

  shape = ALLOC_N(int,1);
  shape[0] = abs(cm);
  ne = ssl2_getary(e,1,shape);
  free(shape);
  shape = ALLOC_N(int,2);
  shape[0] = n;
  shape[1] = abs(cm)+2;
  nev = ssl2_getary(cev,2,shape);
  free(shape);

  free(e);
  free(cev);

  if ( icon!=0 )
    fprintf(stderr,"icon is %d\n",icon);
  return rb_ary_new3(3,ne,nev,INT2NUM(it));
}


static VALUE
rb_teig1(self, d, sd)
     VALUE self, d, sd;
{
  float *cd;
  float *csd;
  int n;
  float *e;
  float *ev;
  int m;
  int icon;

  VALUE ne;
  VALUE nev;
  int *shape;

  int nn;

  cd = ssl2_getcary(d,&nn);
  n = nn;
  if ( n<1 )
    rb_raise(rb_eRuntimeError, "d.length must be >= !");

  csd = ssl2_getcary(sd,&nn);
  if ( nn!=n )
    rb_raise(rb_eRuntimeError, "sd.length != d.length");

  e = ALLOC_N(float,n); 
  ev = ALLOC_N(float,n*n); 

  teig1_(cd, csd, &n, e, ev, &n, &m, &icon);
  free(cd);
  free(csd);

  shape = ALLOC_N(int,1);
  shape[0] = n;
  ne = ssl2_getary(e,1,shape);
  free(shape);
  shape = ALLOC_N(int,2);
  shape[0] = n;
  shape[1] = n;
  nev = ssl2_getary(ev,2,shape);
  free(shape);

  free(e);
  free(ev);

  if ( icon!=0 )
    fprintf(stderr,"icon is %d\n",icon);
  return rb_ary_new3(3,ne,nev,INT2NUM(m));
}


static VALUE
rb_teig2(self, d, sd, m)
     VALUE self, d, sd, m;
{
  float *cd;
  float *csd;
  int n;
  int cm;
  float *e;
  float *ev;
  float *vw;
  int icon;

  VALUE ne;
  VALUE nev;
  int *shape;

  int nn;

  cm = NUM2INT(m);

  cd = ssl2_getcary(d,&nn);
  n = nn;
  if ( n<1 )
    rb_raise(rb_eRuntimeError, "d.length must be >= !");

  csd = ssl2_getcary(sd,&nn);
  if ( nn!=n )
    rb_raise(rb_eRuntimeError, "sd.length != d.length");

  if ( abs(cm)<1 )
    rb_raise(rb_eRuntimeError, "abs(m) must be >= 1");
  if ( abs(cm)>n )
    rb_raise(rb_eRuntimeError, "abs(m) must be <= n");

  e = ALLOC_N(float,n); 
  ev = ALLOC_N(float,n*n); 
  vw = ALLOC_N(float,5*n);

  teig2_(cd, csd, &n, &cm, e, ev, &n, vw, &icon);
  free(vw);
  free(cd);
  free(csd);

  shape = ALLOC_N(int,1);
  shape[0] = n;
  ne = ssl2_getary(e,1,shape);
  free(shape);
  shape = ALLOC_N(int,2);
  shape[0] = n;
  shape[1] = n;
  nev = ssl2_getary(ev,2,shape);
  free(shape);

  free(e);
  free(ev);

  if ( icon!=0 )
    fprintf(stderr,"icon is %d\n",icon);
  return rb_ary_new3(2,ne,nev);
}


static VALUE
rb_gseg2(self, a, b, m, epsz, epst)
     VALUE self, a, b, m, epsz, epst;
{
  float *ca;
  float *cb;
  int n;
  int cm;
  float cepsz;
  float cepst;
  float *e;
  float *ev;
  float *vw;
  int icon;

  VALUE ne;
  VALUE nev;
  int *shape;

  int nn;

  cm = NUM2INT(m);
  cepsz = (float) NUM2DBL(epsz);
  cepst = (float) NUM2DBL(epst);

  ca = ssl2_getcary(a,&nn);
  n = (sqrt(8*nn+1)-1)/2;
  if ( nn!=n*(n+1)/2 )
    rb_raise(rb_eRuntimeError, "a.length must be = n*(n+1)/2");

  cb = ssl2_getcary(b,&nn);
  if ( nn!=n*(n+1)/2 )
    rb_raise(rb_eRuntimeError, "b.length must be = n*(n+1)/2");

  if ( abs(cm)>n )
    rb_raise(rb_eRuntimeError, "abs(m) must be <= n");

  e = ALLOC_N(float,abs(cm)); 
  ev = ALLOC_N(float,n*abs(cm)); 
  vw = ALLOC_N(float,7*n);

  gseg2_(ca, cb, &n, &cm, &cepsz, &cepst, e, ev, &n, vw, &icon);
  free(vw);
  free(ca);
  free(cb);

  shape = ALLOC_N(int,1);
  shape[0] = abs(cm);
  ne = ssl2_getary(e,1,shape);
  free(shape);
  shape = ALLOC_N(int,2);
  shape[0] = n;
  shape[1] = abs(cm);
  nev = ssl2_getary(ev,2,shape);
  free(shape);

  free(e);
  free(ev);

  if ( icon!=0 )
    fprintf(stderr,"icon is %d\n",icon);
  return rb_ary_new3(2,ne,nev);
}


static VALUE
rb_gbseg(self, a, b, nh, m, epsz, epst, lm, ev)
     VALUE self, a, b, nh, m, epsz, epst, lm, ev;
{
  float *ca;
  float *cb;
  int n;
  int cnh;
  int cm;
  float cepsz;
  float cepst;
  int clm;
  float *e;
  float *cev;
  int it;
  float *vw;
  int icon;

  VALUE ne;
  VALUE nev;
  int *shape;

  int nn;

  cnh = NUM2INT(nh);
  cm = NUM2INT(m);
  clm = NUM2INT(lm);
  cepsz = (float) NUM2DBL(epsz);
  cepst = (float) NUM2DBL(epst);

  ca = ssl2_getcary(a,&nn);
  n = nn/(cnh+1)+cnh/2;
  if ( nn!=n*(cnh+1)-cnh*(cnh+1)/2 )
    rb_raise(rb_eRuntimeError, "a.length must be = n*(nh+1)-nh*(nh+1)/2");

  cb = ssl2_getcary(b,&nn);
  if ( nn!=n*(cnh+1)-cnh*(cnh+1)/2 )
    rb_raise(rb_eRuntimeError, "b.length must be = n*(nh+1)-nh*(nh+1)/2");

  cev = ssl2_getcary(ev,&nn);
  if ( nn!=n*cm )
    rb_raise(rb_eRuntimeError, "ev.length must be = n*m");

  if ( abs(cm)>n )
    rb_raise(rb_eRuntimeError, "abs(m) must be <= n");

  if ( cnh<0 )
    rb_raise(rb_eRuntimeError, "nh must be >= 0");
  if ( cnh>=n )
    rb_raise(rb_eRuntimeError, "nh must be < n");

  if ( clm==0 )
    rb_raise(rb_eRuntimeError, "lm must be >= 0");

  e = ALLOC_N(float,abs(cm)); 
  REALLOC_N(cev,float,n*(abs(cm)+2)); 
  vw = ALLOC_N(float,2*n+cm*(3*cm+1)/2);

  gbseg_(ca, cb, &n, &cnh, &cm, &cepsz, &cepst, &clm, e, cev, &n, &it, vw, &icon);
  free(vw);
  free(ca);
  free(cb);

  shape = ALLOC_N(int,1);
  shape[0] = abs(cm);
  ne = ssl2_getary(e,1,shape);
  free(shape);
  shape = ALLOC_N(int,2);
  shape[0] = abs(n);
  shape[1] = abs(cm)+2;
  nev = ssl2_getary(cev,2,shape);
  free(shape);

  free(e);
  free(cev);

  if ( icon!=0 )
    fprintf(stderr,"icon is %d\n",icon);
  return rb_ary_new3(3,ne,nev,INT2NUM(it));
}


void init_eigenvalue_eigenvector(mSSL2)
     VALUE mSSL2;
{
  rb_define_module_function(mSSL2, "eig1", rb_eig1, -1);
  rb_define_module_function(mSSL2, "ceig2", rb_ceig2, -1);
  rb_define_module_function(mSSL2, "seig1", rb_seig1, 1);
  rb_define_module_function(mSSL2, "seig2", rb_seig2, 2);
  rb_define_module_function(mSSL2, "heig2", rb_heig2, 2);
  rb_define_module_function(mSSL2, "bseg", rb_bseg, 5);
  rb_define_module_function(mSSL2, "bsegj", rb_bsegj, 6);
  rb_define_module_function(mSSL2, "teig1", rb_teig1, 2);
  rb_define_module_function(mSSL2, "teig2", rb_teig2, 3);
  rb_define_module_function(mSSL2, "gseg2", rb_gseg2, 5);
  rb_define_module_function(mSSL2, "gbseg", rb_gbseg, 8);
}
